KEYWORDS = ["OR", "NOT", "AND"]


class SQLConditional:
    def __init__(self, expr):
        self.expr = expr

    def AND(self, other):
        if type(self) != type(other):
            raise TypeError("Not an SQLCondition")
        if other._is_compound:
            expr = self.expr + f" AND ({other.expr})"
        else:
            expr = self.expr + f" AND {other.expr}"
        return type(self)(expr)

    def OR(self, other):
        if type(self) != type(other):
            raise TypeError("Not an SQLCondition")
        if other._is_compound:
            expr = self.expr + f" OR ({other.expr})"
        else:
            expr = self.expr + f" OR {other.expr}"
        return type(self)(expr)

    def NOT(self):
        return type(self)("NOT " + self.expr)

    @property
    def _is_compound(self):
        for x in self.expr.split():
            if x in KEYWORDS:
                return True
        return False

    def __str__(self):
        return self.expr

    def __invert__(self):  # Not
        return self.NOT()

    def __and__(self, other):  # And
        return self.AND(other)

    def __or__(self, other):  # Or
        return self.OR(other)

    __repr__ = __str__



def enforce_type(element, _type=str, exact=False):
    """
    Check if given `element` is an instance of `_type`.
    In case of `exact` checking, subclassing isn't allowed.
    """
    if not exact and isinstance(element, _type):
        return
    if exact and type(element) == _type:
        return

    raise TypeError(f"{element} has to be type {_type.__name__}. Is type {type(element)}")


def enforce_types(iterable, _type=str, exact=False, inner=True):
    """
    Make sure that every element and sub-element of
    given `iterable` is an instance of `_type`.
    In case of `exact` checking, subclassing isn't allowed.
    If `inner` is false, don't check sub-elements
    """
    for elem in iterable:
        try:
            enforce_type(elem, _type, exact)
        except TypeError as exc:
            if inner and hasattr(elem, "__iter__"):
                enforce_types(elem, _type, exact, inner)
            else:
                raise exc


def join(iterable, joiner=" "):
    return joiner.join(part for part in iterable if part != "")



from easysql.raw_queries.common.raw_sql import (  # noqa
    _RawSQLSelect,
    _RawSQLSelectDistinct,
    _RawSQLInsert,
    _RawSQLUpdate,
    _RawSQLDelete,
    _RawSQLWhere,
    _RawSQLLike,
    _RawSQLBetween,
    _RawSQLIn,
    _RawSQLOrder,
    _RawSQLTable,
)



import easysql.raw_queries.common as common
from easysql.queries import Query


def wrap(raw_class):
    def construct(*args, **kwargs):
        return raw_class(*args, **kwargs)

    return construct


_METHODS = {
    "postgres": {
        "select": wrap(common._RawSQLSelect),
        "selectdistinct": wrap(common._RawSQLSelectDistinct),
        "insert": wrap(common._RawSQLInsert),
        "update": wrap(common._RawSQLUpdate),
        "delete": wrap(common._RawSQLDelete),
        "where": wrap(common._RawSQLWhere),
        "like": wrap(common._RawSQLLike),
        "between": wrap(common._RawSQLBetween),
        "in": wrap(common._RawSQLIn),
        "order": wrap(common._RawSQLOrder),
        "table": wrap(common._RawSQLTable),
    },
    "mysql": {
        "select": wrap(common._RawSQLSelect),
        "selectdistinct": wrap(common._RawSQLSelectDistinct),
        "insert": wrap(common._RawSQLInsert),
        "update": wrap(common._RawSQLUpdate),
        "delete": wrap(common._RawSQLDelete),
        "where": wrap(common._RawSQLWhere),
        "like": wrap(common._RawSQLLike),
        "between": wrap(common._RawSQLBetween),
        "in": wrap(common._RawSQLIn),
        "order": wrap(common._RawSQLOrder),
        "table": wrap(common._RawSQLTable),
    },
}


class DBType:
    def __init__(self, name):
        if name.lower() not in _METHODS.keys():
            raise ValueError(f"{name} is not a valid DB name")
        self._name = name.lower()

    def __getattr__(self, attr):
        if attr in _METHODS[self._name].keys():
            return _METHODS[self._name][attr]
        else:
            return super().__getattribute__(attr)

    def query(self, *args, **kwargs):
        return Query(*args, **kwargs, db=self)



from easysql.util import enforce_types, join


class _RawSQL:
    @property
    def query(self):
        return "<query string>"

    def __str__(self):
        return self.query


class _RawSQLSelect(_RawSQL):
    def __init__(self, column_names, table_name, constraints=None):
        if constraints is None:
            constraints = []

        enforce_types([table_name, column_names], str)

        self.columns = column_names
        self.table = table_name
        self.constraints = constraints

    @property
    def _select(self):
        sql_columns = ", ".join(self.columns)
        return f"SELECT {sql_columns}"

    @property
    def _from(self):
        return f"FROM {self.table}"

    @property
    def _constraints(self):
        return "\n".join(where.query for where in self.constraints)

    @property
    def query(self):
        return join([self._select, self._from, self._constraints]) + ";"


class _RawSQLSelectDistinct(_RawSQLSelect):
    @property
    def _select(self):
        sql_columns = ", ".join(self.columns)
        return f"SELECT DISTINCT {sql_columns}"


class _RawSQLSelectTop(_RawSQLSelect):
    def __init__(self, row_count, columns, table_name, constraints=None):
        if not constraints:
            constraints = []
        self.rows = row_count
        super().__init__(columns, table_name, constraints)

    def _select(self):
        return f"SELECT TOP {self.rows} {self.columns}"


class _RawSQLInsert(_RawSQL):
    def __init__(self, table_name, values, column_names=""):
        enforce_types([values, table_name, column_names])

        self.values = values
        self.table = table_name
        self.columns = column_names

    @property
    def _insert(self):
        insert = f"INSERT INTO {self.table}"
        if self.columns:
            insert += f" ({', '.join(self.columns)})"
        return insert

    @property
    def _values(self):
        sql_values = ", ".join(self.values)
        return f"VALUES ({sql_values})"

    @property
    def query(self):
        return join([self._insert, self._values]) + ";"


class _RawSQLInsertIntoSelect(_RawSQLInsert):
    def __init__(self, table_name, select_query):
        self.table = table_name
        self.select_query = select_query

    @property
    def _insert(self):
        return f"INSERT INTO {self.table}"

    @property
    def _select(self):
        return self.select_query

    @property
    def query(self):
        return join([self._insert, self._select])


class _RawSQLUpdate(_RawSQL):
    def __init__(self, table_name, column_value_map: dict, constraints=None):
        if constraints is None:
            constraints = []

        enforce_types([column_value_map.items(), table_name], str)

        self.column_value_map = column_value_map
        self.table = table_name
        self.constraints = constraints

    @property
    def _update(self):
        return f"UPDATE {self.table}"

    @property
    def _set(self):
        sql_col_val_map = ", ".join(" = ".join(item) for item in self.column_value_map.items())
        return f"SET {sql_col_val_map}"

    @property
    def _constraints(self):
        return join([where.query for where in self.constraints], "\n")

    @property
    def query(self):
        return join([self._update, self._set, self._constraints]) + ";"


class _RawSQLDelete(_RawSQL):
    def __init__(self, table_name, constraints=None):
        if constraints is None:
            constraints = []

        enforce_types([table_name])
        self.table = table_name
        self.constraints = constraints

    @property
    def _delete(self):
        return f"DELETE FROM {self.table}"

    @property
    def _constraints(self):
        return join([where.query for where in self.constraints], "\n")

    @property
    def query(self):
        return (
            join(
                [self._delete] + [self._constraints]
                if self._constraints != ""
                else [self._delete]
            ) + ";"
        )


class _RawSQLWhere(_RawSQL):
    def __init__(self, condition, extra=""):
        enforce_types([condition], str)
        self.condition = condition
        self.extra = extra

    @property
    def _extra(self):
        return str(self.extra)

    @property
    def _where(self):
        return f"WHERE {self.condition}"

    @property
    def query(self):
        return join([self._where, self._extra])


class _RawSQLLike(_RawSQL):
    def __init__(self, pattern):
        enforce_types([pattern])
        self.pattern = pattern

    @property
    def _like(self):
        return f"LIKE {self.pattern}"

    @property
    def query(self):
        return self._like


class _RawSQLBetween(_RawSQL):
    def __init__(self, num1: str, num2: str):
        enforce_types([num1, num2])
        self.num1 = num1
        self.num2 = num2

    @property
    def _between(self):
        return f"BETWEEN {self.num1} AND {self.num2}"

    @property
    def query(self):
        return self._between


class _RawConditionsInput(_RawSQL):
    def __init__(self, *conditions, keyword=None):
        enforce_types([[str(cond) for cond in conditions], keyword], str)
        self._kw = keyword
        self.conditions = [
            str(cond) if type(cond) != list else cond for cond in conditions
        ]

    @property
    def query(self):
        return (" " + self._kw + " ").upper().join(self.conditions)


class _RawSQLIn(_RawConditionsInput):
    def __init__(self, values):
        super().__init__(values, keyword="in")

    @property
    def query(self):
        sql_conditions = ", ".join(*self.conditions)
        return f"{self._kw.upper()} ({sql_conditions})"


class _RawSQLOrder(_RawSQL):
    def __init__(self, value, asc):
        self.asc = asc
        self.value = value

    @property
    def _order(self):
        return "ORDER"

    @property
    def _by(self):
        return f"BY {self.value}"

    @property
    def _asc(self):
        return "ASC" if self.asc else "DESC"

    @property
    def query(self):
        return join([self._order, self._by, self._asc])


class _RawSQLTable(_RawSQL):
    def __init__(self, name, column_value_map: dict, if_not_exists=True):
        self.name = name
        self.map = column_value_map
        self.if_not_exists = if_not_exists

    @property
    def _create(self):
        return f"CREATE TABLE {'IF NOT EXISTS ' if self.if_not_exists else ''}{self.name}"

    @property
    def _values(self):
        return f"({', '.join([' '.join(items) for items in self.map.items()])})"

    @property
    def query(self):
        return join([self._create, self._values]) + ";"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="easysql",
    version="1.0.1",
    author="TeamNightSky",
    description="Python SQL query generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TeamNightSky/EasySQL",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: SQL",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Database",
    ],
    python_requires=">=3.6",
)



from easysql.queries.conditions import SQLConditional as sqc


class Query:
    def __init__(self, method, *args, _filter=None, db):
        self._args = args
        self._method = method
        self._filter = _filter
        self.db = db

        _class = getattr(self.db, self._method)
        if _class.__name__ not in ["_insert", "_table"] and self._filter is not None:
            self.method = _class(*self._args, [self._filter])
        else:
            self.method = _class(*self._args)

    def __repr__(self):
        method_name = self._method.upper().strip("_")
        return f"<SQLQuery object '{method_name}'>"

    def __str__(self):
        return self.string

    def __lshift__(self, other):
        if type(other).__name__ != "SQLConditional":
            raise TypeError(f"Cannot add condition to Query with type {type(other)}")
        if not self._filter:
            self._filter = self.Where(other)._filter
            return
        self._filter = self.Where(sqc(self._filter.condition).AND(other))._filter

    @property
    def string(self):
        return str(self.method)

    def OrderBy(self, value, asc=True):
        order = self.db.order(value, asc)
        return Query(self._method, db=self.db, *self._args, _filter=order)

    # Filters
    def Where(self, condition=""):
        where = self.db.where(str(condition))
        where.extra = ""
        return Query(self._method, db=self.db, *self._args, _filter=where)

    def Between(self, bottom, top):
        self._filter.extra += str(self.db.between(str(bottom), str(top)))
        return self

    def Like(self, pattern):
        self._filter.extra += str(self.db.like(pattern))
        return self

    def In(self, values):
        self._filter.extra += str(
            getattr(self.db, "in")([str(value) for value in values])
        )
        return self
